@@ -1,6 +1,6 @@
 #!/usr/bin/perl
 
-# Copyright 2005-2007 Messiah College. All rights reserved.
+# Copyright 2005-2009 Messiah College. All rights reserved.
 # Jason Long <jlong@messiah.edu>
 
 # Copyright (c) 2004 Anthony D. Urso. All rights reserved.
@@ -43,6 +43,19 @@ Mail::DKIM::Verifier - verifies a DKIM-signed message
   # what is the result of the verify?
   my $result = $dkim->result;
 
+  # there might be multiple signatures, what is the result per signature?
+  foreach my $signature ($dkim->signatures)
+  {
+      print "signature identity: " . $signature->identity . "\n";
+      print "verify result: " . $signature->result_detail . "\n";
+  }
+
+  # the alleged author of the email may specify how to handle email
+  foreach my $policy ($dkim->policies)
+  {
+      die "fraudulent message" if ($policy->apply($dkim) eq "reject");
+  }
+
 =cut
 
 =head1 DESCRIPTION
@@ -54,7 +67,7 @@ message has been completely read, the signatures are verified and the
 results of the verification can be accessed.
 
 To use the verifier, first create the verifier object. Then start
-"feeding" it the email message to be verified. When all the headers
+"feeding" it the email message to be verified. When all the _headers_
 have been read, the verifier:
 
  1. checks whether any DomainKeys/DKIM signatures were found
@@ -62,12 +75,21 @@ have been read, the verifier:
  3. sets up the appropriate algorithms and canonicalization objects
  4. canonicalizes the headers and computes the header hash
 
-Then, when the body of the message has been completely fed into the
+Then, when the _body_ of the message has been completely fed into the
 verifier, the body hash is computed and the signatures are verified.
 
+The results of the verification can be checked with L</"result()">
+or L</"signatures()">.
+
+Messages that do not verify may be checked against the alleged sender's
+published signing policy with L</"policies()"> and
+L<Mail::DKIM::Policy/"apply()">.
+
 =head1 CONSTRUCTOR
 
-=head2 new() - construct an object-oriented verifier
+=head2 new()
+
+Constructs an object-oriented verifier.
 
   my $dkim = Mail::DKIM::Verifier->new();
 
@@ -89,7 +111,8 @@ is written to the referenced string or file handle.
 package Mail::DKIM::Verifier;
 use base "Mail::DKIM::Common";
 use Carp;
-our $VERSION = 0.33;
+our $VERSION = 0.44;
+our $MAX_SIGNATURES_TO_PROCESS = 50;
 
 sub init
 {
@@ -174,6 +197,10 @@ sub add_signature
 	croak "wrong number of arguments" unless (@_ == 1);
 	my ($signature) = @_;
 
+	# ignore signature headers once we've seen 50 or so
+	# this protects against abuse.
+	return if (@{$self->{signatures}} > $MAX_SIGNATURES_TO_PROCESS);
+
 	push @{$self->{signatures}}, $signature;
 
 	unless ($self->check_signature($signature))
@@ -183,6 +210,9 @@ sub add_signature
 		return;
 	}
 
+	# signature looks ok, go ahead and query for the public key
+	$signature->fetch_public_key;
+
 	# create a canonicalization filter and algorithm
 	my $algorithm_class = $signature->get_algorithm_class(
 				$signature->algorithm);
@@ -267,16 +297,21 @@ sub check_signature
 		return 0;
 	}
 
-	unless ($signature->domain ne '')
+	unless (defined $signature->domain)
 	{
 		# no domain specified
-		$self->{signature_reject_reason} =
-			!defined($signature->domain) ? "missing d tag"
-			: "invalid domain in d tag";
+		$self->{signature_reject_reason} = "missing d tag";
 		return 0;
 	}
 
-	unless ($signature->selector)
+	if ($signature->domain eq '')
+	{
+		# blank domain
+		$self->{signature_reject_reason} = "invalid domain in d tag";
+		return 0;
+	}
+
+	unless (defined $signature->selector)
 	{
 		# no selector specified
 		$self->{signature_reject_reason} = "missing s tag";
@@ -369,21 +404,36 @@ sub finish_header
 
 	foreach my $algorithm (@{$self->{algorithms}})
 	{
-		$algorithm->finish_header;
+		$algorithm->finish_header(Headers => $self->{headers});
 	}
 
-	# For each parsed signature, check it for validity. If none are valid,
-	# our result is "invalid" and our result detail will be the reason
-	# why the last signature was invalid.
+	# stop processing signatures that are already known to be invalid
+	@{$self->{algorithms}} = grep
+		{
+			my $sig = $_->signature;
+			!($sig->result && $sig->result eq "invalid");
+		} @{$self->{algorithms}};
 
-	foreach my $signature (@{$self->{signatures}})
+	if (@{$self->{algorithms}} == 0
+		&& @{$self->{signatures}} > 0)
 	{
+		$self->{result} = $self->{signatures}->[0]->result || "invalid";
+		$self->{details} = $self->{signatures}->[0]->{verify_details} || $self->{signature_reject_reason};
+		return;
+	}
+}
+
+sub _check_and_verify_signature
+{
+	my $self = shift;
+	my ($algorithm) = @_;
+
+		# check signature
+		my $signature = $algorithm->signature;
 		unless (check_signature_identity($signature))
 		{
 			$self->{signature_reject_reason} = "bad identity";
-			$signature->result("invalid",
-				$self->{signature_reject_reason});
-			next;
+			return ("invalid", $self->{signature_reject_reason});
 		}
 
 		# get public key
@@ -397,43 +447,13 @@ sub finish_header
 			my $E = $@;
 			chomp $E;
 			$self->{signature_reject_reason} = "public key: $E";
-			$signature->result("invalid",
-				$self->{signature_reject_reason});
-			next;
+			return ("invalid", $self->{signature_reject_reason});
 		}
 
 		unless ($self->check_public_key($signature, $pkey))
 		{
-			$signature->result("invalid",
-				$self->{signature_reject_reason});
-			next;
+			return ("invalid", $self->{signature_reject_reason});
 		}
-	}
-
-	# stop processing signatures that are already known to be invalid
-	@{$self->{algorithms}} = grep
-		{
-			my $sig = $_->signature;
-			!($sig->result && $sig->result eq "invalid");
-		} @{$self->{algorithms}};
-
-	if (@{$self->{algorithms}} == 0
-		&& @{$self->{signatures}} > 0)
-	{
-		$self->{result} = $self->{signatures}->[0]->result;
-		$self->{details} = $self->{signatures}->[0]->{verify_details};
-		return;
-	}
-}
-
-sub finish_body
-{
-	my $self = shift;
-
-	foreach my $algorithm (@{$self->{algorithms}})
-	{
-		# finish canonicalizing
-		$algorithm->finish_body;
 
 		# verify signature
 		my $result;
@@ -459,6 +479,19 @@ sub finish_body
 			$result = "fail";
 			$details = $E;
 		}
+	return ($result, $details);
+}
+
+sub finish_body
+{
+	my $self = shift;
+
+	foreach my $algorithm (@{$self->{algorithms}})
+	{
+		# finish canonicalizing
+		$algorithm->finish_body;
+
+		my ($result, $details) = $self->_check_and_verify_signature($algorithm);
 
 		# save the results of this signature verification
 		$algorithm->{result} = $result;
@@ -478,7 +511,9 @@ sub finish_body
 
 =head1 METHODS
 
-=head2 PRINT() - feed part of the message to the verifier
+=head2 PRINT()
+
+Feeds part of the message to the verifier.
 
   $dkim->PRINT("a line of the message\015\012");
   $dkim->PRINT("more of");
@@ -488,31 +523,78 @@ Feeds content of the message being verified into the verifier.
 The API is designed this way so that the entire message does NOT need
 to be read into memory at once.
 
-=head2 CLOSE() - call this when finished feeding in the message
+Please note that although the PRINT() method expects you to use
+SMTP-style line termination characters, you should NOT use the
+SMTP-style dot-stuffing technique described in RFC 2821 section 4.5.2.
+Nor should you use a <CR><LF>.<CR><LF> sequence to terminate the
+message.
+
+=head2 CLOSE()
+
+Call this when finished feeding in the message.
 
   $dkim->CLOSE;
 
 This method finishes the canonicalization process, computes a hash,
 and verifies the signature.
 
-=head2 fetch_author_policy() - retrieves a signing policy from DNS
+=head2 fetch_author_domain_policies()
+
+Retrieves ADSP records from DNS.
+
+  my @policies = $dkim->fetch_author_domain_policies;
+  foreach my $policy (@policies)
+  {
+      my $policy_result = $policy->apply($dkim);
+  }
+
+This method will retrieve all applicable
+"author-domain-signing-practices" published in DNS for this message.
+Author policies are keyed to the email address(es) in the From: header,
+i.e. the claimed author of the message.
+
+This method returns a *list* of policy records, since there is allowed
+to be zero or multiple email addresses in the From: header.
+
+The result of the apply() method is one of: "accept", "reject", "neutral".
+
+See also: L</"policies()">.
+
+=cut
+
+sub fetch_author_domain_policies
+{
+	my $self = shift;
+	use Mail::DKIM::AuthorDomainPolicy;
+
+	return () unless $self->{headers_by_name}->{from};
+	my @list = Mail::Address->parse(
+		$self->{headers_by_name}->{from}
+		);
+	my @authors = map { $_->address } @list;
+
+	# fetch the policies
+	return map {
+		Mail::DKIM::AuthorDomainPolicy->fetch(
+			Protocol => "dns",
+			Author => $_,
+			)
+		} @authors;
+}
+
+=head2 fetch_author_policy()
+
+Retrieves a signing policy from DNS.
 
   my $policy = $dkim->fetch_author_policy;
   my $policy_result = $policy->apply($dkim);
 
-The "author" policy, as I call it, is the DKIM Sender Signing Practices
+This method retrieves the DKIM Sender Signing Practices
 record as described in Internet Draft draft-ietf-dkim-ssp-00-01dc.
-I call it the "author" policy because it is keyed to the email address
-in the From: header, i.e. the author of the message.
+This Internet Draft is now obsolete; this method is only kept for
+backward-compatibility purposes.
 
-The IETF is still actively working on this Internet Draft, so the
-exact mechanisms are subject to change.
-
-If the email being verified has no From header at all
-(which violates email standards),
-then this method will C<die>.
-
-The result of the apply() method is one of: "accept", "reject", "neutral".
+Please use the L</"policies()"> method instead.
 
 =cut
 
@@ -532,7 +614,9 @@ sub fetch_author_policy
 			);
 }
 
-=head2 fetch_sender_policy() - retrieves a signing policy from DNS
+=head2 fetch_sender_policy()
+
+Retrieves a signing policy from DNS.
 
   my $policy = $dkim->fetch_sender_policy;
   my $policy_result = $policy->apply($dkim);
@@ -550,26 +634,30 @@ then this method will C<die>.
 
 The result of the apply() method is one of: "accept", "reject", "neutral".
 
+See also: L</"policies()">.
+
 =cut
 
 sub fetch_sender_policy
 {
 	my $self = shift;
-	use Mail::DKIM::Policy;
+	use Mail::DKIM::DkPolicy;
 
 	# determine addresses found in the "From" and "Sender" headers
 	my $author = $self->message_originator->address;
 	my $sender = $self->message_sender->address;
 
 	# fetch the policy
-	return Mail::DKIM::Policy->fetch(
+	return Mail::DKIM::DkPolicy->fetch(
 			Protocol => "dns",
 			Author => $author,
 			Sender => $sender,
 			);
 }
 
-=head2 load() - load the entire message from a file handle
+=head2 load()
+
+Load the entire message from a file handle.
 
   $dkim->load($file_handle);
 
@@ -577,7 +665,9 @@ Reads a complete message from the designated file handle,
 feeding it into the verifier. The message must use <CRLF> line
 terminators (same as the SMTP protocol).
 
-=head2 message_originator() - access the "From" header
+=head2 message_originator()
+
+Access the "From" header.
 
   my $address = $dkim->message_originator;
 
@@ -593,7 +683,9 @@ To get just the email address part, do:
 
 See also L</"message_sender()">.
 
-=head2 message_sender() - access the "From" or "Sender" header
+=head2 message_sender()
+
+Access the "From" or "Sender" header.
 
   my $address = $dkim->message_sender;
 
@@ -612,8 +704,50 @@ transmission of the message. For example, if a secretary were to send a
 message for another person, the "sender" would be the secretary and
 the "originator" would be the actual author.
 
+=head2 policies()
 
-=head2 result() - access the result of the verification
+Retrieves applicable signing policies from DNS.
+
+  my @policies = $dkim->policies;
+  foreach my $policy (@policies)
+  {
+      $policy_result = $policy->apply($dkim);
+      # $policy_result is one of "accept", "reject", "neutral"
+  }
+
+This method searches for and returns any signing policies that would
+apply to this message. Signing policies are selected based on the
+domain that the message *claims* to be from. So, for example, if
+a message claims to be from security@bank, and forwarded by
+trusted@listserv, when in reality the message came from foe@evilcorp,
+this method would check for signing policies for security@bank and
+trusted@listserv. The signing policies might tell whether
+foe@evilcorp (the real sender) is allowed to send mail claiming
+to be from your bank or your listserv.
+
+I say "might tell", because in reality this is still really hard to
+specify with any accuracy. In addition, most senders do not publish
+useful policies.
+
+=cut
+
+sub policies
+{
+	my $self = shift;
+
+	my $sender_policy = eval { $self->fetch_sender_policy() };
+	my $author_policy = eval { $self->fetch_author_policy() };
+	return (
+		$sender_policy ? $sender_policy : (),
+		$author_policy ? $author_policy : (),
+		$self->fetch_author_domain_policies(),
+		);
+}
+
+
+=head2 result()
+
+Access the result of the verification.
 
   my $result = $dkim->result;
 
@@ -634,7 +768,15 @@ does not contain a correct value for the message.
 =item invalid
 
 Returned if a DKIM-Signature could not be checked because of a problem
-in the signature itself or the public key record.
+in the signature itself or the public key record. I.e. the signature
+could not be processed.
+
+=item temperror
+
+Returned if a DKIM-Signature could not be checked due to some error
+which is likely transient in nature, such as a temporary inability
+to retrieve a public key. A later attempt may produce a better
+result.
 
 =item none
 
@@ -644,14 +786,19 @@ Returned if no DKIM-Signature headers (valid or invalid) were found.
 
 In case of multiple signatures, the "best" result will be returned.
 Best is defined as "pass", followed by "fail", "invalid", and "none".
+To examine the results of individual signatures, use the L</"signatures()">
+method to retrieve the signature objects. See
+L<Mail::DKIM::Signature/"result()">.
 
 =cut
 
-=head2 result_detail() - access the result, plus details if available
+=head2 result_detail()
+
+Access the result, plus details if available.
 
   my $detail = $dkim->result_detail;
 
-The detail is constructed by taking the result (i.e. one of "pass", "fail",
+The detail is constructed by taking the result (e.g. "pass", "fail",
 "invalid" or "none") and appending any details provided by the verification
 process in parenthesis.
 
@@ -688,7 +835,9 @@ The following are possible results from the result_detail() method:
   invalid (public key: OpenSSL error: ...)
   none
 
-=head2 signature() - access the message's DKIM signature
+=head2 signature()
+
+Access the message's DKIM signature.
 
   my $sig = $dkim->signature;
 
@@ -701,7 +850,9 @@ Best is defined as "pass", followed by "fail", "invalid", and "none".
 
 =cut
 
-=head2 signatures() - access all of this message's signatures
+=head2 signatures()
+
+Access all of this message's signatures.
 
   my @all_signatures = $dkim->signatures;
 
@@ -723,7 +874,7 @@ Jason Long, E<lt>jlong@messiah.eduE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2006-2007 by Messiah College
+Copyright (C) 2006-2009 by Messiah College
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.8.6 or,
